/*
 * Ok, this is the stable release, after a long code overhaul. We can now read
 * as many files as the user throws at us, generate as many random pads to mix
 * in as the user asks for, and do it all much cleaner.
 * Future needs:
 * - have it check input filenames in the pad-md5-xxxxx.dat format to see if
 *   their md5 sum really matches what their name claims it does.
 * - have an option to generate pads alone.
 * - make file pointer arrays expand to use whatever memory is needed.
 *
 * If you can improve upon this in any way, or have some kind of program that
 * uses it, please let me know. I'm interested.
 *
 * Everything contained in this file is subject to the terms of the GPL lisence
 * which is available in the COPYING file in this directory.
 *
 * --xercist@mindless.com
 *
 */


#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <openssl/rand.h>
#include "pad.h"
#include "md5.h"

int main(argc, argv)
  int argc;
  char **argv;
{
  char *ifname[256], *ofname[256];                      /* Names of files */
  FILE *ifp[256], *ofp[256];                  /* file pointers themselves */
  struct md5_ctx context[256];                            /* md5 contexts */
  unsigned char digest[256][16];                           /* md5 digests */
  unsigned char currentbyte, readbyte;   /* just a place to put new bytes */
  char *outputname=NULL;                              /* output file name */
  char buf[256];                                /* temporary string space */
  int defined_output=0, lastout, rands=-1, i, j;             /* misc ints */
  unsigned long bytecount=0, bytelimit=0; /* to limit/expand the filesize */

  /* Let's parse some command params, eh? */
  j=0;
  ifname[0]=NULL;
  for (i=1; (i < argc); i++)
  {
    if (argv[i][0] == '-')
    {
      switch (argv[i][1])
      {
        case 'h':
	  printversion();
          printhelp();
          exit(0);
        break;
	case 'v':
	  printversion();
	  exit(0);
	break;
	case 'r':
	  i++;
	  if (argc == i)
	  {
	    fprintf(stderr, "-r must be used with a number.\n\n");
	    printhelp();
	    exit(1);
	  }
	  if ((!(rands = atoi(argv[i]))) || (rands < 0))
	  {
	    fprintf(stderr, "-r must be used with a positive integer.\n\n");
	    printhelp();
	    exit(1);
	  }
	break;
	case 's':
	  i++;
	  if (argc == i)
	  {
	    fprintf(stderr, "-s must be used with a filesize\n\n");
	    printhelp();
	    exit(1);
	  }
	  if (!(bytelimit = bytes_in(argv[i])))
	  {
	    fprintf(stderr, "Error parsing output size\n\n");
	    printhelp();
	    exit(1);
	  }
	break;
	case 'o':
	  i++;
	  if (argc == i)
	  {
	    fprintf(stderr, "-o must be used with an output filename.\n\n");
	    printhelp();
	    exit(1);
	  }
	  outputname=argv[i];
	  defined_output=1;
	break;
	default:
          fprintf(stderr, "Unknown option %s\n\n", argv[i]);
	  printhelp();
	  exit(1);
      }
    } else {
      if ((j+rands) > 254)
      {
        /* This limitation will be removed as soon as I get around to making
         * the file pointer arrays self-expanding */
	fprintf(stderr, "Random pads + Input files cannot exceed 255.\n");
	exit(1);
      } else {
        ifname[j] = argv[i];
        ifname[j+1] = NULL;
        j++;
      }
    }
  }
  if (j == 0)
  {
    printversion();
    printhelp();
    exit(1);
  }

  /* Let's open all our input files, and figure out how many
   * random pads should be mixed in, that is, if the user didn't tell us */
  ifp[0]=NULL;
  for (i=0; (ifname[i] != NULL); i++)
  {
    if ((ifp[i]=fopen(ifname[i], "r"))==NULL)
    {
      fprintf(stderr,"Error opening %s: %s\n", ifname[i], strerror(errno));
      exit(1);
    }
    ifp[i+1] = NULL;
    if ((rands == -1) && (i > 0))
      rands=0;
  }
  if (rands == -1)
    rands=1;

  for (i=0; ifp[i] != NULL; i++)
  {
    if ((j != filesize(ifp[i])) && (i != 0))
    {
      fprintf(stderr, "Warning: Input files sizes differ. Output will match the size of the smallest input.\n");
      break;
    }
    j=filesize(ifp[i]);
  }

  /* And now to figure out the names of our output files... */
  ofname[0] = NULL;
  for (i=0; i<rands; i++)
  {
    ofname[i] = tempnam("./", "pad"); /* if tmpnam doesn't like "./" and starts
					 giving us /tmp/, this could be a
					 problem later on when renaming */
    ofname[i+1] = NULL;
  }

  if (defined_output)
    ofname[i] = outputname;
  else
    ofname[i] = tempnam("./", "pad");
  lastout=i;
  ofname[i+1] = NULL;


  /* Open the output files for writing */
  for (i=0; ofname[i] != NULL; i++)
  {
    if ((ofp[i]=fopen(ofname[i], "w"))==NULL)
    {
      fprintf(stderr,"Error opening %s: %s\n", ofname[i], strerror(errno));
      exit(1);
    }
    ofp[i+1] = NULL;
  }

  for (i=0; ofp[i] != NULL; i++)
    MD5_start(&context[i]); /* start calculating md5 for each output file */
  
  /* This while() is the heart of the program - the read/write loop */
  while (1) 
  {
    currentbyte='\0';
    for (i=0; ifp[i] != NULL; i++)
    {
      readbyte=fgetc(ifp[i]);
      currentbyte=(currentbyte^readbyte);
      if (file_end(ifp))
	goto endfile; /* yes, I hate goto, but this is the best way... */
      for (j=0; j<rands; j++)
      {
        RAND_bytes(&readbyte, 1);
        MD5_feed(&context[j], readbyte);
        fputc(readbyte, ofp[j]);
        currentbyte=currentbyte^readbyte;
      }
    }
    MD5_feed(&context[lastout], currentbyte);
    fputc(currentbyte, ofp[lastout]);
    bytecount++;
    if ((bytelimit && (bytecount == bytelimit)))
    { /* user's -s flag limited us to this size - we're done */
      fprintf(stderr, "Input not EOF but size limit hit, clipping\n");
      break;
    }
  }
endfile:

  if ((bytelimit) && (bytecount < bytelimit))
  { /* write random bytes to the file/s until size is met */
    fprintf(stderr, "Input EOF with %lu byte%s to go - filling in with random data\n", (bytelimit-bytecount), ((bytelimit-bytecount)>1)?"s":"");
    while (bytecount < bytelimit)
    {
      for (j=0; (j<rands+1); j++)
      {
        RAND_bytes(&readbyte, 1);
        MD5_feed(&context[j], readbyte);
        fputc(readbyte, ofp[j]);
      }
      bytecount++;
    }
  }

  /* Close all outputs, tell MD5 processor they're finished */
  for (i=0; ofp[i] != NULL; i++)
  {
    fclose(ofp[i]);
    MD5_stop(&context[i], digest[i]);
  }

  /* Close all inputs */
  for (i=0; ifp[i] != NULL; i++)
  {
    fclose(ifp[i]);
  }

  /* Rename output files to match their md5-sums */
  for (i=0; ofname[i] != NULL; i++)
  {
    if ((defined_output) && (lastout == i))
      sprintf(buf, outputname);
    else
      sprintf(buf, "pad-md5-%s.dat", hexize(digest[i]));

    if (rename(ofname[i], buf)) /* note: no renaming across filesystems */
      fprintf(stderr, "Error renaming: %s\n", strerror(errno));
    else 
      fprintf(stderr, "Wrote %s\n", buf);
  }
  exit(0);
}

/* Return size of fp passed */
long filesize(FILE *fp)
{
  struct stat buf;
  if (fstat(fileno(fp), &buf))
  {
    return -1;
  } else {
    return buf.st_size;
  }
}

/* print out the help message */
void printhelp()
{
  fprintf(stderr, "Usage: pad [options] [input files]

options:

  -r [num]    - Pad will XOR in [num] pads generated from random data. if this
                is ommited, pad will use either 1 or 0 random pads, depending on
                if there is only one input file or more than one, respectively.

  -o [output] - The result of all the XOR operations will be stored in this
                file. If [output] is ommitted, pad will name the file according
                to its MD5-sum, along with the rest of the random pads.

  -s [size]   - The output data will be [size] bytes. If this is less than the
                smallest input file, the output will be clipped to match, if it
                is larger, random data will be appended to the end.

  -h          - Show usage.

");
}

void printversion()
{
  fprintf(stderr, "Pad v%s by xercist\n", VERSION);
}

/* return 1 if one of the input files has hit eof */
int file_end(FILE *ifp[256])
{
  int i;
  for (i=0; ifp[i] != NULL; i++) 
  {
    if (feof(ifp[i]))
      return 1;
  }
  return 0;
}

/* convert digest to readable string */
char *hexize(unsigned char digest[16])
{
  const char hexdigits[17] = "0123456789abcdef";
  char buf[32];
  int j;

  buf[0] = '\0';
  for ( j=0 ; j<16 ; j++ )
  {
    sprintf(buf, "%s%c", buf, (hexdigits[digest[j]>>4]));
    sprintf(buf, "%s%c", buf, (hexdigits[digest[j]&0xf]));
  }
  return((char *)strdup(buf));
}

/* take a string like "5k" and return a number like 5120 */
unsigned long bytes_in(char *string)
{
  int h, i, length=0;
  unsigned long total=0, multiplier=1;
  for (h=0;h<2;h++)
  {
    for (i=0;i<strlen(string);i++)
    {
      if (isdigit(string[i]))
      {
	if (!h) {
	  length++;
	} else {
	  if ((total+((string[i]-'0')*pow(10,(length-(i+1))))*multiplier) > 4294967294u)
	    return 0;
          total += ((string[i]-'0')*pow(10,(length-(i+1))));
        }
      } else {
	if (!h)
	{
	  switch (string[i])
	  {
	    case 'b': case 'B':
	      multiplier=1;
	    break;
	    case 'k': case 'K':
	      multiplier=1024;
	    break;
	    case 'm': case 'M':
	      multiplier=pow(1024,2);
            break;
	    default:
	      return 0;
	    break;
	  }
	  break;
      }	else
	  break;
      }
    }
  }
  total = total*multiplier;
  return total;
}

